pub enum Error {
CompactStakeOverflow,
CompactTargetOverflow,
CompactInvalidIndex,
ArithmeticError(&'static str),
InvalidSupportEdge,
}Expand description
The errors that might occur in the this crate and compact.
Variants
CompactStakeOverflow
While going from compact to staked, the stake of all the edges has gone above the total and the last stake cannot be assigned.
CompactTargetOverflow
The compact type has a voter who’s number of targets is out of bound.
CompactInvalidIndex
One of the index functions returned none.
ArithmeticError(&'static str)
An error occurred in some arithmetic operation.
InvalidSupportEdge
The data provided to create support map was invalid.